@@ -10,6 +10,7 @@ import torch.multiprocessing
 from pathlib import Path
 from typing import Dict, List, Optional
 
+from azureml.core import Run
 from pytorch_lightning import Trainer, seed_everything
 
 from health_azure import AzureRunInfo
@@ -18,7 +19,7 @@ from health_azure.utils import (create_run_recovery_id, ENV_OMPI_COMM_WORLD_RANK
                                 is_running_in_azure_ml, PARENT_RUN_CONTEXT, RUN_CONTEXT,
                                 aggregate_hyperdrive_metrics, get_metrics_for_childless_run,
                                 ENV_GLOBAL_RANK, ENV_LOCAL_RANK, ENV_NODE_RANK,
-                                is_local_rank_zero, is_global_rank_zero,)
+                                is_local_rank_zero, is_global_rank_zero, create_aml_run_object)
 
 from health_ml.experiment_config import ExperimentConfig
 from health_ml.lightning_container import LightningContainer
@@ -78,6 +79,7 @@ class MLRunner:
                                                     project_root=self.project_root,
                                                     run_context=RUN_CONTEXT)
         self.trainer: Optional[Trainer] = None
+        self.azureml_run_for_logging: Optional[Run] = None
 
     def set_run_tags_from_parent(self) -> None:
         """
@@ -176,21 +178,43 @@ class MLRunner:
         # Set random seeds just before training. Ensure that dataloader workers are also seeded correctly.
         seed_everything(self.container.get_effective_random_seed(), workers=True)
 
-        # get the container's datamodule
+        # Get the container's datamodule
         self.data_module = self.container.get_data_module()
 
-        checkpoint_path_for_recovery = self.checkpoint_handler.get_recovery_or_checkpoint_path_train()
+        # Create an AzureML run for logging if running outside AzureML. This run will be used for metrics logging
+        # during both training and inference. We can't rely on the automatically generated run inside the AzureMLLogger
+        # class because two of those logger objects will be created, so training and inference metrics would be logged
+        # in different runs.
+        if self.container.log_from_vm:
+            run = create_aml_run_object(experiment_name=self.container.effective_experiment_name)
+            # Display name should already be set when creating the Run object, but in some scenarios this
+            # does not happen. Hence, set it again.
+            run.display_name = self.container.tag if self.container.tag else None
+            self.azureml_run_for_logging = run
 
-        self.trainer, self.storing_logger = create_lightning_trainer(
-            self.container, checkpoint_path_for_recovery,
-            num_nodes=self.container.num_nodes,
-            multiple_trainloader_mode=self.get_multiple_trainloader_mode())
+        if not self.container.run_inference_only:
 
-        rank_info = ", ".join(f"{env}: {os.getenv(env)}" for env in [ENV_GLOBAL_RANK, ENV_LOCAL_RANK, ENV_NODE_RANK])
-        logging.info(f"Environment variables: {rank_info}. trainer.global_rank: {self.trainer.global_rank}")
+            checkpoint_path_for_recovery = self.checkpoint_handler.get_recovery_or_checkpoint_path_train()
+            if not checkpoint_path_for_recovery and self.container.src_checkpoint:
+                # If there is no recovery checkpoint (e.g job hasn't been resubmitted) and a source checkpoint is given,
+                # use it to resume training.
+                checkpoint_path_for_recovery = self.checkpoint_handler.trained_weights_path
 
-    def load_model_checkpoint_after_training(self) -> None:
-        self.checkpoint_handler.additional_training_done()
+            self.trainer, self.storing_logger = create_lightning_trainer(
+                container=self.container,
+                resume_from_checkpoint=checkpoint_path_for_recovery,
+                num_nodes=self.container.num_nodes,
+                multiple_trainloader_mode=self.get_multiple_trainloader_mode(),
+                azureml_run_for_logging=self.azureml_run_for_logging)
+
+            rank_info = ", ".join(
+                f"{env}: {os.getenv(env)}" for env in [ENV_GLOBAL_RANK, ENV_LOCAL_RANK, ENV_NODE_RANK]
+            )
+            logging.info(f"Environment variables: {rank_info}. trainer.global_rank: {self.trainer.global_rank}")
+
+    def load_model_checkpoint(self) -> None:
+        if not self.container.run_inference_only:
+            self.checkpoint_handler.additional_training_done()
         checkpoint_path_for_inference = self.checkpoint_handler.get_checkpoint_to_test()
         self.container.load_model_checkpoint(checkpoint_path_for_inference)
 
@@ -243,8 +267,9 @@ class MLRunner:
             assert self.trainer, "Trainer should be initialized before training. Call self.init_training() first."
             self.trainer.fit(self.container.model, datamodule=self.data_module)
 
-        assert self.trainer.logger is not None
-        self.trainer.logger.finalize('success')
+        for logger in self.trainer.loggers:
+            assert logger is not None
+            logger.finalize('success')
 
     def run_validation(self) -> None:
         """
@@ -270,7 +295,15 @@ class MLRunner:
             # uneven inputs.
             self.container.max_num_gpus = 1
 
-            trainer, _ = create_lightning_trainer(self.container, num_nodes=1)
+            checkpoint_path = (
+                self.checkpoint_handler.get_checkpoint_to_test() if self.container.src_checkpoint else None
+            )
+            trainer, _ = create_lightning_trainer(
+                container=self.container,
+                resume_from_checkpoint=checkpoint_path,
+                num_nodes=1,
+                azureml_run_for_logging=self.azureml_run_for_logging
+            )
 
             # Change to the outputs folder so that the model can write to current working directory, and still
             # everything is put into the right place in AzureML (there, only the contents of the "outputs" folder
@@ -323,28 +356,37 @@ class MLRunner:
         Driver function to run a ML experiment
         """
         self.setup()
-        self.init_training()
-        # Backup the environment variables in case we need to run a second training in the unit tests.
-        old_environ = dict(os.environ)
+        try:
+            self.init_training()
+            if not self.container.run_inference_only:
+                # Backup the environment variables in case we need to run a second training in the unit tests.
+                old_environ = dict(os.environ)
 
-        # do training
-        with logging_section("Model training"):
-            self.run_training()
+                # do training
+                with logging_section("Model training"):
+                    self.run_training()
 
-        # load model checkpoint for custom inference or additional validation step
-        if self.container.has_custom_test_step() or self.container.run_extra_val_epoch:
-            self.load_model_checkpoint_after_training()
+                # load model checkpoint for custom inference or additional validation step
+                if self.container.has_custom_test_step() or self.container.run_extra_val_epoch:
+                    self.load_model_checkpoint()
 
-        # Run extra validation epoch if enabled
-        if self.container.run_extra_val_epoch:
-            with logging_section("Model Validation to save plots on validation set"):
-                self.run_validation()
+                # Run extra validation epoch if enabled
+                if self.container.run_extra_val_epoch:
+                    with logging_section("Model Validation to save plots on validation set"):
+                        self.run_validation()
 
-        # Kill all processes besides rank 0
-        self.after_ddp_cleanup(old_environ)
+                # Kill all processes besides rank 0
+                self.after_ddp_cleanup(old_environ)
 
-        # Run inference on a single device
-        with logging_section("Model inference"):
-            self.run_inference()
+            # Run inference on a single device
+            with logging_section("Model inference"):
+                self.run_inference()
 
-        self.run_regression_test()
+            self.run_regression_test()
+
+        finally:
+            if self.azureml_run_for_logging is not None:
+                try:
+                    self.azureml_run_for_logging.complete()
+                except Exception as ex:
+                    logging.error("Failed to complete AzureML run: %s", ex)
